from __future__ import absolute_import

import doubleratchet

class RootChain(doubleratchet.kdfchains.KDFChain):
    def __init__(self, key = None):
        super(RootChain, self).__init__(
            doubleratchet.recommended.RootKeyKDF(
                "SHA-256",
                "WhisperRatchet".encode("US-ASCII")
            ),
            key
        )

    def serialize(self):
        return {
            "super": super(RootChain, self).serialize()
        }

    @classmethod
    def fromSerialized(cls, serialized, *args, **kwargs):
        return super(RootChain, cls).fromSerialized(
            serialized["super"],
            *args,
            **kwargs
        )



from __future__ import absolute_import

import doubleratchet

from .sendrecvchain import SendRecvChain

class SymmetricKeyRatchet(doubleratchet.ratchets.SymmetricKeyRatchet):
    def __init__(self):
        super(SymmetricKeyRatchet, self).__init__(
            SendRecvChain,
            SendRecvChain
        )

    def serialize(self):
        return {
            "super": super(SymmetricKeyRatchet, self).serialize()
        }

    @classmethod
    def fromSerialized(cls, serialized, *args, **kwargs):
        return super(SymmetricKeyRatchet, cls).fromSerialized(
            serialized["super"],
            *args,
            **kwargs
        )



from __future__ import absolute_import

import x3dh

class PublicKeyEncoder(x3dh.PublicKeyEncoder):
    @staticmethod
    def encodePublicKey(key, key_type):
        if key_type == "25519":
            return b"\x05" + key

        raise NotImplementedError



from __future__ import absolute_import

import doubleratchet

class SendRecvChain(doubleratchet.kdfchains.ConstKDFChain):
    def __init__(self, key = None):
        """
        Note: The key parameter MUST NOT be None.
        """
        super(SendRecvChain, self).__init__(
            doubleratchet.recommended.ChainKeyKDF("SHA-256"),
            None,
            key
        )

    def serialize(self):
        return {
            "super": super(SendRecvChain, self).serialize()
        }

    @classmethod
    def fromSerialized(cls, serialized, *args, **kwargs):
        return super(SendRecvChain, cls).fromSerialized(
            serialized["super"],
            *args,
            **kwargs
        )



from __future__ import absolute_import

from omemo.backends import Backend

from .doubleratchet import DoubleRatchet
from .wireformat import WireFormat
from .x3dh import State as X3DHState

BACKEND = Backend(WireFormat, X3DHState, DoubleRatchet)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: WhisperTextProtocol.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='WhisperTextProtocol.proto',
  package='',
  syntax='proto2',
  serialized_pb=_b('\n\x19WhisperTextProtocol.proto\"R\n\rSignalMessage\x12\x16\n\x0e\x64h_ratchet_key\x18\x01 \x01(\x0c\x12\t\n\x01n\x18\x02 \x01(\r\x12\n\n\x02pn\x18\x03 \x01(\r\x12\x12\n\nciphertext\x18\x04 \x01(\x0c\"\x7f\n\x13PreKeySignalMessage\x12\x17\n\x0fregistration_id\x18\x05 \x01(\r\x12\x0f\n\x07otpk_id\x18\x01 \x01(\r\x12\x0e\n\x06spk_id\x18\x06 \x01(\r\x12\n\n\x02\x65k\x18\x02 \x01(\x0c\x12\n\n\x02ik\x18\x03 \x01(\x0c\x12\x16\n\x0esignal_message\x18\x04 \x01(\x0c')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_SIGNALMESSAGE = _descriptor.Descriptor(
  name='SignalMessage',
  full_name='SignalMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dh_ratchet_key', full_name='SignalMessage.dh_ratchet_key', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='n', full_name='SignalMessage.n', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='pn', full_name='SignalMessage.pn', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ciphertext', full_name='SignalMessage.ciphertext', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=29,
  serialized_end=111,
)


_PREKEYSIGNALMESSAGE = _descriptor.Descriptor(
  name='PreKeySignalMessage',
  full_name='PreKeySignalMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='registration_id', full_name='PreKeySignalMessage.registration_id', index=0,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='otpk_id', full_name='PreKeySignalMessage.otpk_id', index=1,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='spk_id', full_name='PreKeySignalMessage.spk_id', index=2,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ek', full_name='PreKeySignalMessage.ek', index=3,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ik', full_name='PreKeySignalMessage.ik', index=4,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='signal_message', full_name='PreKeySignalMessage.signal_message', index=5,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=113,
  serialized_end=240,
)

DESCRIPTOR.message_types_by_name['SignalMessage'] = _SIGNALMESSAGE
DESCRIPTOR.message_types_by_name['PreKeySignalMessage'] = _PREKEYSIGNALMESSAGE

SignalMessage = _reflection.GeneratedProtocolMessageType('SignalMessage', (_message.Message,), dict(
  DESCRIPTOR = _SIGNALMESSAGE,
  __module__ = 'WhisperTextProtocol_pb2'
  # @@protoc_insertion_point(class_scope:SignalMessage)
  ))
_sym_db.RegisterMessage(SignalMessage)

PreKeySignalMessage = _reflection.GeneratedProtocolMessageType('PreKeySignalMessage', (_message.Message,), dict(
  DESCRIPTOR = _PREKEYSIGNALMESSAGE,
  __module__ = 'WhisperTextProtocol_pb2'
  # @@protoc_insertion_point(class_scope:PreKeySignalMessage)
  ))
_sym_db.RegisterMessage(PreKeySignalMessage)


# @@protoc_insertion_point(module_scope)



from __future__ import absolute_import

import base64

import doubleratchet
import x3dh

from .cbcaead import CBCAEAD
from .rootchain import RootChain
from .symmetrickeyratchet import SymmetricKeyRatchet

class DoubleRatchet(doubleratchet.ratchets.DoubleRatchet):
    def __init__(
        self,
        ad,
        root_key   = None,
        own_key    = None,
        other_pub  = None,
        skr        = None, # ONLY USED FOR DESERIALIZATION
        root_chain = None  # ONLY USED FOR DESERIALIZATION
    ):
        if (root_key == None) == (root_chain == None):
            raise ValueError(
                "Exactly one of root_key, root_chain must be set! Not both, not neither."
            )

        self.__ad = ad

        if skr == None:
            self.__skr = SymmetricKeyRatchet()
        else:
            self.__skr = skr

        if root_chain == None:
            self.__root_chain = RootChain(root_key)
        else:
            self.__root_chain = root_chain

        super(DoubleRatchet, self).__init__(
            self.__skr, # symmetric_key_ratchet
            CBCAEAD(),  # aead
            self.__ad,  # ad
            5000,       # message_key_store_max
            self.__root_chain, # root_chain
            x3dh.implementations.KeyPairCurve25519, # encryption_key_pair_class
            own_key,
            other_pub
        )

    def serialize(self):
        ad = {
            "IK_own"   : base64.b64encode(self.__ad["IK_own"]).decode("US-ASCII"),
            "IK_other" : base64.b64encode(self.__ad["IK_other"]).decode("US-ASCII")
        }

        return {
            "super"      : super(DoubleRatchet, self).serialize(),
            "ad"         : ad,
            "skr"        : self.__skr.serialize(),
            "root_chain" : self.__root_chain.serialize()
        }

    @classmethod
    def fromSerialized(cls, serialized):
        ad = serialized["ad"]

        ad = {
            "IK_own"   : base64.b64decode(ad["IK_own"].encode("US-ASCII")),
            "IK_other" : base64.b64decode(ad["IK_other"].encode("US-ASCII"))
        }

        return super(DoubleRatchet, cls).fromSerialized(
            serialized["super"],
            ad,
            skr = SymmetricKeyRatchet.fromSerialized(serialized["skr"]),
            root_chain = RootChain.fromSerialized(serialized["root_chain"])
        )

    def _makeAD(self, header, ad):
        return ad

    def encryptMessage(self, *args, **kwargs):
        result = super(DoubleRatchet, self).encryptMessage(*args, **kwargs)
        result["additional"] = result["ciphertext"]["additional"]
        result["ciphertext"] = result["ciphertext"]["ciphertext"]
        return result



from __future__ import absolute_import

import doubleratchet

from omemo.backends import WireFormat as WF
from omemo.exceptions import WireFormatException

from . import whispertextprotocol_pb2 as wtp

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, hmac

CURRENT_MAJOR_VERSION = 3
CURRENT_MINOR_VERSION = 3
KEY_TYPE_25519 = 5
MAC_SIZE = 8

CRYPTOGRAPHY_BACKEND = default_backend()

def calculateMAC(data, key, IK_sender, IK_receiver):
    global CRYPTOGRAPHY_BACKEND

    # Build the authentication
    auth = hmac.HMAC(
        key,
        hashes.SHA256(),
        backend = CRYPTOGRAPHY_BACKEND
    )

    auth.update(IK_sender + IK_receiver + data)

    # Return the part of the mac that is used my libsignal
    return auth.finalize()[:MAC_SIZE]

def checkVersion(data):
    try:
        version = ord(data[0])
    except TypeError:
        version = data[0]

    major_version = (version >> 4) & 0x0F
    minor_version = (version >> 0) & 0x0F

    if major_version < CURRENT_MAJOR_VERSION or minor_version < CURRENT_MINOR_VERSION:
        raise WireFormatException("Legacy version detected.")

    if major_version > CURRENT_MAJOR_VERSION or minor_version > CURRENT_MINOR_VERSION:
        raise WireFormatException("Newer/unknown version detected.")

    return data[1:]

def prependVersion(data):
    return bytes(bytearray([ CURRENT_MAJOR_VERSION << 4 | CURRENT_MINOR_VERSION ])) + data

def decodePublicKey(key):
    if len(key) != 33:
        raise WireFormatException("The key field must contain 33 bytes of data.")

    try:
        key_type = ord(key[0])
    except TypeError:
        key_type = key[0]

    if key_type != KEY_TYPE_25519:
        raise WireFormatException("Unknown key type.")

    return key[1:]

def encodePublicKey(key):
    return bytes(bytearray([ KEY_TYPE_25519 ])) + key

class WireFormat(WF):
    @staticmethod
    def messageFromWire(obj):
        # Due to the nature the mac is calculated by signal, the authentication
        # verification has to be done later in an additional step.

        # Remove the mac
        mac = obj[-MAC_SIZE:]
        obj = obj[:-MAC_SIZE]

        # Check and remove the version
        obj = checkVersion(obj)

        # Unpack the protobuf structure
        obj = wtp.SignalMessage.FromString(obj)

        if not (
            obj.HasField("dh_ratchet_key") and
            obj.HasField("n") and
            obj.HasField("ciphertext")
        ):
            raise WireFormatException("Message incomplete.")

        return {
            "ciphertext": obj.ciphertext,
            "header": doubleratchet.Header(
                decodePublicKey(obj.dh_ratchet_key),
                obj.n,
                obj.pn
            ),
            "additional": mac
        }

    @staticmethod
    def finalizeMessageFromWire(obj, additional):
        dr_additional = additional["DoubleRatchet"]

        ad  = dr_additional["ad"]
        key = dr_additional["key"]
        mac = calculateMAC(obj[:-MAC_SIZE], key, ad["IK_other"], ad["IK_own"])

        if not additional["WireFormat"] == mac:
            raise WireFormatException("Message authentication failed.")

    @staticmethod
    def messageToWire(ciphertext, header, additional):
        # Build the protobuf structure
        wire = wtp.SignalMessage()
        wire.ciphertext = ciphertext
        wire.n = header.n

        if header.pn:
            wire.pn = header.pn
        else:
            wire.pn = 0

        wire.dh_ratchet_key = encodePublicKey(header.dh_pub)
        data = wire.SerializeToString()

        # Prepend the message version
        data = prependVersion(data)

        # The specification of the DoubleRatchet protocol recommends to calculate the mac
        # of ad+ciphertext and append the result to the ciphertext.

        # WhisperSystems instead calculate the mac of ad + the whole protobuf encoded
        # message and truncate the mac to 8 bytes.
        #
        # This way the whole message is authenticated and not only the ciphertext.
        # (idk about the truncation though).
        dr_additional = additional["DoubleRatchet"]

        ad  = dr_additional["ad"]
        key = dr_additional["key"]

        data += calculateMAC(data, key, ad["IK_own"], ad["IK_other"])
        
        return data

    @staticmethod
    def preKeyMessageFromWire(obj):
        obj = checkVersion(obj)

        obj = wtp.PreKeySignalMessage.FromString(obj)

        if not (
            obj.HasField("spk_id") and
            obj.HasField("ek") and
            obj.HasField("ik") and
            obj.HasField("signal_message") and
            obj.HasField("otpk_id")
        ):
            raise WireFormatException("Pre key message incomplete.")

        return {
            "session_init_data": {
                "registration_id": obj.registration_id,
                "otpk_id": obj.otpk_id,
                "spk_id": obj.spk_id,
                "ek": decodePublicKey(obj.ek),
                "ik": decodePublicKey(obj.ik)
            },
            "message": obj.signal_message,
            "additional": None
        }

    @staticmethod
    def finalizePreKeyMessageFromWire(obj, additional):
        # TODO: Verify the mac of the contained message
        pass

    @staticmethod
    def preKeyMessageToWire(session_init_data, message, additional):
        wire = wtp.PreKeySignalMessage()
        wire.registration_id = 0 # This parameter has no use in OMEMO.
        wire.otpk_id = session_init_data["otpk_id"]
        wire.spk_id  = session_init_data["spk_id"]
        wire.ek = encodePublicKey(session_init_data["ek"])
        wire.ik = encodePublicKey(session_init_data["ik"])
        wire.signal_message = message
        data = wire.SerializeToString()
        
        return prependVersion(data)



from __future__ import absolute_import

from .publickeyencoder import PublicKeyEncoder

import x3dh

class State(x3dh.State):
    def __init__(self):
        """
        Sets the constructor parameters to the defaults used by OMEMO.
        The curve, min_num_otpks and max_num_otpks parameters were found in the XEP
        (https://xmpp.org/extensions/xep-0384.html).

        The hash_function and info_string parameters were found in the source code of
        libsignal-protocol-java
        (https://github.com/WhisperSystems/libsignal-protocol-java).

        The timeout for the SPK is defaulted to one week.
        """

        return super(State, self).__init__(
            "WhisperText".encode("US-ASCII"), # info_string
            "25519",                          # curve
            "SHA-256",                        # hash_function
            7 * 24 * 60 * 60,                 # spk_timeout
            20,                               # min_num_otpks
            100,                              # max_num_otpks
            PublicKeyEncoder                  # public_key_encoder_class
        )

    def serialize(self):
        return {
            "super": super(State, self).serialize()
        }

    @classmethod
    def fromSerialized(cls, serialized, *args, **kwargs):
        return super(State, cls).fromSerialized(
            serialized["super"],
            *args,
            **kwargs
        )

    def getSharedSecretActive(self, *args, **kwargs):
        # X3DH is specified to build the associated data as follows: IK_A || IK_B.
        # As per usual, WhisperSystems weren't satisfied with their own solution and
        # instead of using the ad as built by X3DH they ALWAYS do:
        #     IK_sender || IK_receiver.
        # That means, when decyprint a message, another ad is used as when encrypting a
        # message.
        #
        # To allow for this to happen, we split the ad returned by X3DH into IK_own and
        # IK_other.

        result = super(State, self).getSharedSecretActive(*args, **kwargs)

        result["ad"] = {
            "IK_own"   : result["ad"][:33],
            "IK_other" : result["ad"][33:]
        }

        return result

    def getSharedSecretPassive(self, *args, **kwargs):
        result = super(State, self).getSharedSecretPassive(*args, **kwargs)

        # See getSharedSecretActive for an explanation
        result["ad"] = {
            "IK_own"   : result["ad"][33:],
            "IK_other" : result["ad"][:33]
        }

        return result
